import importlib.util
import pkgutil
from dataclasses import dataclass, field
from types import ModuleType

from .transform_df import Transform


@dataclass
class Pipeline:
    _transforms: list[Transform] = field(default_factory=list)

    def discover_transforms(self, *modules: ModuleType):
        
        for mod in modules:
            
            if not isinstance(mod, ModuleType):
                raise TypeError("modules must be of type ModuleType")
            submodules = pkgutil.iter_modules(mod.__path__)
            self._discover_transforms_path(mod.__path__)
            
            
    def _discover_transforms_path(self, path:list[str]):
        submodules = pkgutil.iter_modules(path)
        for submod in submodules:
            mod_path = submod.module_finder.path + "/" + submod.name

            if submod.ispkg:
                self._discover_transforms_path([mod_path])
            else:

                spec = importlib.util.spec_from_file_location('test', mod_path + ".py")
                module = importlib.util.module_from_spec(spec)  # type: ignore
                # sys.modules[module_name] = module  # Register the module in sys.modules
                spec.loader.exec_module(  # type:ignore
                    module
                )  # Execute the module in its own namespace
                
                
                for item in module.__dict__.values():
                    if isinstance(item, Transform):
                        self._transforms.append(item)
                pass        

    


from dataclasses import dataclass

from foundry_dev_tools import FoundryContext
from foundry_dev_tools.errors.dataset import BranchNotFoundError
from foundry_dev_tools.utils.caches.spark_caches import (
    _infer_dataset_format,
    _validate_cache_key,
)
from pyspark.sql import DataFrame, SparkSession
from asyncer import asyncify
from transforms.runner.data_source.base import (
    BranchNotFoundError as BranchNotFoundErrorBase,
)
from transforms.runner.data_source.base import DataSource


@dataclass
class FoundrySource(DataSource):
    ctx: FoundryContext
    session: SparkSession

    async def download_dataset(self, dataset_path_or_rid: str, branch: str) -> DataFrame:
        try:
            dataset_identity = await asyncify(self.ctx.cached_foundry_client._get_dataset_identity)(
                dataset_path_or_rid, branch
            )
            if dataset_identity.get("last_transaction") is None:
                raise BranchNotFoundErrorBase("FOUNDRY")

            last_path, dataset_identity = await asyncify(self.ctx.cached_foundry_client.fetch_dataset)(
                dataset_path_or_rid, branch
            )
            self.last_path = last_path
            _validate_cache_key(dataset_identity)
            try:
                inferred_format = _infer_dataset_format(
                    self.ctx.cached_foundry_client.cache.get_cache_dir(),
                    dataset_identity,
                )
                path = self.ctx.cached_foundry_client.cache._get_storage_location(
                    dataset_identity, inferred_format
                )
                if inferred_format == "parquet":
                    try:
                        return self.session.read.parquet(str(path.joinpath("*.parquet")))
                    except Exception as e:
                        
                        # Exception for partitioned parquet files - mostly for duckdb compat
                        return self.session.read.parquet(str(path.joinpath("**", "*.parquet")))
                elif inferred_format == "csv":
                    ds = self.ctx.get_dataset(dataset_path_or_rid)
                    df: "DataFrame" = ds._context.foundry_sql_server.query_foundry_sql(
                        f"select * from `{dataset_path_or_rid}`",
                        branch=branch,
                        return_type="spark",
                    ) 
                    return df
                raise NotImplementedError(f"Format {inferred_format} is not supported")

            except FileNotFoundError as exc:
                msg = f"{dataset_identity}"
                raise KeyError(msg) from exc
        except BranchNotFoundError:
            print(
                f"[FOUNDRY] Branch [{branch}] not found for dataset [{dataset_path_or_rid}]"
            )
            raise BranchNotFoundErrorBase("FOUNDRY")

    async def download_for_branches(self, dataset_path_or_rid: str, branches: list[str]):
        for branch in branches:
            return await self.download_dataset(dataset_path_or_rid, branch=branch)

        raise BranchNotFoundErrorBase("FOUNDRY")

    def get_last_transaction(
        self, dataset_path_or_rid: str, branches: list[str]
    ) -> DataFrame:
        raise NotImplementedError()



from .colexpects import all, any, col, primary_key, schema
from .count_expectations import count
from .grouped import group_by

__all__ = [
    "count",
    "schema",
    "primary_key",
    "col",
    "any",
    "all",
    "group_by",
]



from contextlib import contextmanager
from pathlib import Path
from typing import Any, TypedDict

import duckdb
from foundry_dev_tools import Config, FoundryContext, JWTTokenProvider
from foundry_dev_tools.errors.dataset import BranchNotFoundError
from foundry_dev_tools.utils import api_types
from pytest import MonkeyPatch

from transforms.api import Input, Output, transform_df
from transforms.manage import FoundryManager


class TestStoreDataset(TypedDict):
    dataset_rid: str
    available_branches: list[str]
    dataset_name: str


dataset_store: list[TestStoreDataset] = [
    {
        "dataset_rid": "rid_1",
        "available_branches": ["master", "dev"],
        "dataset_name": "somefilename1",
    },
    {
        "dataset_rid": "rid_2",
        "available_branches": ["master"],
        "dataset_name": "somefilename2",
    },
]


def mock_dataset_identity(
    monkeypatch: MonkeyPatch,
    obj: Any,
    rids_to_available_branches: list[TestStoreDataset],
):
    def _get_dataset_identity(
        dataset_rid: str, branch: str
    ) -> api_types.DatasetIdentity:
        res = [
            dataset
            for dataset in rids_to_available_branches
            if dataset_rid == dataset['dataset_rid'] and branch in dataset["available_branches"]
        ]
        if len(res) > 0:
            return api_types.DatasetIdentity(
                dataset_path=res[0]['dataset_name'],
                dataset_rid=dataset_rid,
                last_transaction_rid=dataset_rid,
                last_transaction=None,
            )
        raise BranchNotFoundError(info="not found branch")

    monkeypatch.setattr(obj, "_get_dataset_identity", _get_dataset_identity)


def mock_get_temp_files(
    monkeypatch: MonkeyPatch,
    obj: Any,
    rids_to_available_branches: list[TestStoreDataset],
):
    @contextmanager
    def mock_get_temp_files(
        dataset_rid: str,
        view: str,
    ):
        res = [
            dataset
            for dataset in rids_to_available_branches
            if dataset["dataset_rid"] == dataset_rid and view in dataset["available_branches"]
        ]
        if len(res) > 0:
            yield Path(__file__).parent / "test_datasets/iris"
        else:
            raise BranchNotFoundError(info="not found branch")

    monkeypatch.setattr(obj, "download_dataset_files", mock_get_temp_files)


def test_collecting_single_input(monkeypatch: MonkeyPatch):
    mngr = FoundryManager(
        duckdb_conn=duckdb.connect(":memory:"),
        ctx=FoundryContext(
            config=Config(), token_provider=JWTTokenProvider("test", jwt="test2")
        ),
    )

    mock_dataset_identity(monkeypatch, mngr.ctx.cached_foundry_client, dataset_store)
    mock_get_temp_files(monkeypatch, mngr.ctx.cached_foundry_client.api, dataset_store)
    
    
    mngr.get_dataset_from_foundry_into_duckdb("rid_1", "master")

    check_if_dataset_exists_and_has_rows(mngr, "fndry_master.somefilename1")


def test_transform_collection(monkeypatch: MonkeyPatch):
    mngr = FoundryManager(
        duckdb_conn=duckdb.connect(":memory:"),
        ctx=FoundryContext(
            config=Config(), token_provider=JWTTokenProvider("test", jwt="test2")
        ),
    )
    mock_dataset_identity(
        monkeypatch,
        mngr.ctx.cached_foundry_client,
        dataset_store,
    )
    mock_get_temp_files(monkeypatch, mngr.ctx.cached_foundry_client.api, dataset_store)

    from pyspark.sql import DataFrame

    @transform_df(Output("rid_2"), df=Input("rid_1"))
    def transform(df: DataFrame) -> DataFrame:
        return df

    mngr.collect_transform_inputs(transform)
    check_if_dataset_exists_and_has_rows(mngr, "fndry_master.somefilename1")


def check_if_dataset_exists_and_has_rows(mngr: FoundryManager, dataset_name: str):
    query_result: tuple[int] | None = mngr.duckdb_conn.query(
        f"SELECT count(*) FROM {dataset_name}"
    ).fetchone()
    assert query_result is not None
    assert query_result[0] > 0



from typing import Callable, Iterable

from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame


def patch_dataframe_rdd(df: DataFrame):
    import pandas as pd
    from pyspark import RDD
    from pyspark.sql import Row, SparkSession
    from pyspark.sql.types import StructType

    class PatchedRDD:
        def __init__(self, session: SparkSession, pandas_df: PandasDF):
            self.df = pandas_df
            self.num_partitions = 8
            self.session = session

        def mapPartitions(
            self, func: Callable[[Iterable[Row]], Iterable[Row]]
        ) -> "PatchedRDD":
            num_rows = len(self.df)
            partitions = (
                self.df.iloc[i * num_rows : (i + 1) * num_rows].to_dict(
                    orient="records"
                )
                for i in range(self.num_partitions)
            )

            new_dfs: list[PandasDF] = []
            for partition in partitions:
                row_gen = (Row(**row) for row in partition)
                res_data = list(r.asDict() for r in func(row_gen))
                res_df = PandasDF(res_data)
                new_dfs.append(res_df)
            res_df = pd.concat(new_dfs, axis=0)
            new_rdd = PatchedRDD(session=self.session, pandas_df=res_df)
            return new_rdd

        def toDF(self, schema: StructType) -> DataFrame:
            new_data = [Row(**row) for row in self.df.to_dict(orient="records")]
            return self.session.createDataFrame(data=new_data)

    class PatchedDfWithRdd(DataFrame):
        @property
        def rdd(self):
            rdd1 = self.toPandas()
            return PatchedRDD(self.session, rdd1)

    df.__class__ = PatchedDfWithRdd
    return df



from abc import ABC

from pyspark.sql import DataFrame


class DataSink(ABC):
    def __init__(self):
        pass

    def save_transaction(
        self,
        df: DataFrame,
        dataset_path_or_rid: str,
    ) -> None:
        raise NotImplementedError()

    def save_incremental_transaction(
        self,
        df: DataFrame,
        dataset_path_or_rid: str,
        semantic_version: int,
    ):
        # TODO: Implement
        pass





def test_transform():
    from pyspark.sql import DataFrame

    from transforms.api.transform_df import (
        Input,
        Output,
        transform_df,
    )

    @transform_df(output=Output("dataset"), df1=Input("dataset"), df2=Input("dataset"))
    def some_transform(df1: DataFrame, df2: DataFrame):
        return df1
    assert some_transform.inputs["df1"] is not None
    


def test_incremental():
    from pyspark.sql import DataFrame

    from transforms.api.incremental_transform import incremental
    from transforms.api.transform_df import Input, Output, transform_df

    @incremental(require_incremental=True)
    @transform_df(Output("dataset"), df1=Input("dataset"))
    def some_transform(df1: DataFrame):
        return df1
    assert some_transform.inputs["df1"] is not None

def test_all():
    from pyspark.sql import DataFrame

    from transforms.api import configure
    from transforms.api.incremental_transform import incremental
    from transforms.api.transform_df import Input, Output, transform_df

    @configure(profile=["dev"])
    @incremental(require_incremental=True)
    @transform_df(Output("dataset"), df1=Input("dataset"))
    def some_transform(df1: DataFrame):
        return df1
    assert some_transform.inputs["df1"] is not None

def test_multi_output():
    from transforms.api.transform_df import (
        Input,
        Output,
        TransformInput,
        TransformOutput,
        transform,
    )

    
    
    @transform(out1=Output("dataset"), df1=Input("dataset"))
    def some_transform(out1: TransformOutput, df1: TransformInput):
        return df1
    assert some_transform.inputs["df1"] is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Concatenate, Literal, ParamSpec

from pyspark.sql import DataFrame, SparkSession

from transforms.external.systems import Source

if TYPE_CHECKING:
    from .incremental_transform import IncrementalTransformOpts

from .check import Check


@dataclass
class Input:
    path_or_rid: str
    checks: Check | None | list[Check] = None
    branch: str | None = None



class Output:
    def __init__(self, path_or_rid: str, checks: Check | None | list[Check] = None):
        self.path_or_rid = path_or_rid
        if checks is None:
            self.checks = []
        elif isinstance(checks, Check):
            self.checks = [checks]
        else:
            self.checks = checks
        


@dataclass
class Context:
    session: SparkSession
    is_incremental: bool = False


class Transform:
    def __init__(
        self,
        inputs: dict[str, Input],
        outputs: dict[str, Output],
        transform: Callable[..., Any],
        multi_outputs: dict[str, "TransformOutput"] | None = None,
        incremental_opts: "IncrementalTransformOpts | None" = None,
        external_systems: dict[str, "Source"] | None = None,
    ):
        self.inputs = inputs
        self.outputs = outputs
        self.transform = transform
        self.multi_outputs = multi_outputs
        self.incremental_opts = incremental_opts
        self.external_systems = external_systems


DecoratorParamSpec = ParamSpec(
    "DecoratorParamSpec",
)


class TransformInput:
    def __init__(self, df: DataFrame):
        self.df = df

    def dataframe(
        self,
        mode: Literal["current", "previous"] = "current"
    ):
        return self.df


class TransformOutput:
    def __init__(
        self,
        on_dataframe_req: Callable[[Literal["current", "previous"]], DataFrame],
        on_dataframe_write: Callable[[DataFrame, Literal["append", "replace"]], None],
    ):
        self.on_dataframe_req = on_dataframe_req
        self.on_dataframe_write = on_dataframe_write
        self.mode_state: Literal["replace", "append"] = "replace"

    def dataframe(self, mode: Literal["current", "previous"] = "current") -> DataFrame:
        return self.on_dataframe_req(mode)

    def set_mode(self, mode: Literal["append", "replace"]):
        self.mode_state = mode
        return self

    def write_dataframe(self, df: DataFrame):
        return self.on_dataframe_write(df, self.mode_state)


def transform(**kwargs: Input | Output):
    def _transform(transform: Callable[..., Any]):
        inputs: dict[str, Input] = {}
        outputs: dict[str, Output] = {}

        for key, arg in kwargs.items():
            if isinstance(arg, Input):
                inputs[key] = arg

            if isinstance(arg, Output):
                outputs[key] = arg

        def transformed_transform(**kwargs: DataFrame | Source | TransformOutput) -> None:
            new_kwargs: dict[str, TransformInput|Source| TransformOutput] = {}
            for key, value in kwargs.items():
                if isinstance(value, Source):
                    new_kwargs[key] = value
                elif isinstance(value, TransformOutput):
                    new_kwargs[key] = value
                else:
                    new_kwargs[key] = TransformInput(df=value)
            return transform(**new_kwargs)

        return Transform(
            inputs=inputs,
            outputs=outputs,
            transform=transformed_transform,
            multi_outputs={},
        )

    return _transform


def transform_df(output: Output, **kwargs: Input):
    TParams = ParamSpec("TParams")

    def _transform_df(transform: Callable[Concatenate[TParams], DataFrame]):
        return Transform(
            inputs=kwargs,
            outputs={"output": output},
            transform=transform,
            multi_outputs=None,
        )

    return _transform_df



from dataclasses import dataclass

from .transform_df import Transform


@dataclass
class IncrementalTransformOpts:
    semantic_version: int = 1
    snapshot_inputs: list[str] | None = None
    allow_retention: bool = False
    strict_append: bool = False
    require_incremental: bool = False
    v2_semantics: bool = False


class IncrementalInput:
    def __init__(self, path_or_rid: str):
        self.path_or_rid = path_or_rid

    def dataframe(self,):
        pass
        

def incremental(
    require_incremental: bool = False,
    semantic_version: int = 1,
    snapshot_inputs: list[str] | None = None,
    allow_retention: bool = False,
    strict_append: bool = False,
    v2_semantics: bool = False,
):
    def _incremental(transform: Transform):
        if (snapshot_inputs is not None):
            for input in snapshot_inputs:
                if input not in transform.inputs:
                    raise ValueError(
                        f"Input {input} is not defined in transform {transform}"
                    )
        transform.incremental_opts = IncrementalTransformOpts(
            semantic_version=semantic_version,
            snapshot_inputs=snapshot_inputs,
            allow_retention=allow_retention,
            strict_append=strict_append,
            require_incremental=require_incremental,
            v2_semantics=v2_semantics,
        )
        return transform
    return _incremental


from pyspark.sql import SparkSession


def init_sess() -> SparkSession:
    
    spark = (
        SparkSession.builder.master("local[*]")
        .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        .config("spark.sql.execution.arrow.pyspark.fallback.enabled", "true")
        .config("spark.executor.memory", "6g")
        .config("spark.driver.memory", "6g")
        .config("spark.driver.host", "localhost")
        .config("spark.driver.hostname", "localhost")
        .getOrCreate()
    )
    
    return spark


from typing import Any

from transforms.api.transform_df import Transform


def configure(
    profile: str | list[str] | None = None,
    *args: Any,
    **kwargs: Any
):
    def _configure(transform: Transform):
        return transform
    return _configure


from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    import _operator

    from pyspark.sql import Column, DataFrame
    from pyspark.sql import functions as F
import operator as op
from collections.abc import Mapping
from dataclasses import dataclass

from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql.types import DataType

from transforms.expectations.base import Expectation


@dataclass
class SchemaExpectation(Expectation):
    expected_schema: Mapping[str, "DataType"] | DataType

    def run(self, dataframe_to_verify: "DataFrame"):
        if isinstance(self.expected_schema, DataType):
            res = self.expected_schema == dataframe_to_verify.schema
            if not res:
                raise AssertionError(
                    f"Schema of the dataframe is not as expected. Expected: [{self.expected_schema}] Actual: [{dataframe_to_verify.schema}]"
                )
        else:
            diff = set(dataframe_to_verify.columns).difference(self.expected_schema)
            if len(diff) > 0:
                raise AssertionError(
                    f"Schema of the dataframe is not as expected. Missing columns: {diff}"
                )

@dataclass
class SchemaContainsExpectation(Expectation):
    expected_schema: Mapping[str, "DataType"] 

    def run(self, dataframe_to_verify: "DataFrame"):
        diff = set(self.expected_schema).difference(dataframe_to_verify.columns)
        if len(diff) > 0:
            raise AssertionError(
                f"Schema of the dataframe is not as expected. Missing columns: {diff}"
            )

@dataclass
class SchemaBuilder:
    equals = SchemaExpectation
    contains = SchemaContainsExpectation

class PrimaryKeyExpectation(Expectation):
    def __init__(self, *pk: str):
        self.pk = pk

    def run(self, dataframe_to_verify: "DataFrame"):
        from pyspark.sql import functions as F

        res = dataframe_to_verify.select(
            (F.count_distinct(*self.pk) != F.count("*")).alias("result"),
        ).collect()
        if not res[0][0]:
            raise AssertionError("Primary key is not unique")


@dataclass
class ColExpectationIsIn(Expectation):
    col: str
    values_arr: list[str]

    def run(self, dataframe_to_verify: "DataFrame"):
        from pyspark.sql import functions as F

        cnt = dataframe_to_verify.filter(~F.col(self.col).isin(self.values_arr)).count()
        if cnt > 0:
            raise AssertionError(f"Column {self.col} is not in {self.values_arr}")


@dataclass
class ColExpectation(Expectation):
    col: str
    operation: "Callable[[DataFrame], DataFrame]"

    def run(self, dataframe_to_verify: "DataFrame"):
        result = self.operation(dataframe_to_verify)
        res = result.filter(result["result"] == False).limit(10).collect()
        if len(res) > 0:
            raise AssertionError(
                f"Column {self.col} failed to meet expectations, example issues: {res}"
            )


@dataclass
class OpComparisonExpectation(Expectation):
    colname: str
    operator: "Callable[[_operator._SupportsComparison, _operator._SupportsComparison], Column]"
    value: int | float

    def run(self, dataframe_to_verify: "DataFrame"):
        result = dataframe_to_verify.withColumn(
            "result", self.operator(F.col(self.colname), self.value)
        )
        res = result.filter(result["result"] == False).limit(10).collect()
        if len(res) > 0:
            raise AssertionError(
                f"Column {self.colname} failed to meet expectations, example issues: {res}"
            )


@dataclass
class ColExpectationBuilder:
    col: str

    def is_in(self, *values_arr: str) -> Expectation:
        return ColExpectationIsIn(self.col, list(values_arr))

    def rlike(self, pattern: str) -> Expectation:
        def operation(df: DataFrame) -> DataFrame:
            return df.withColumn("result", F.col(self.col).rlike(pattern))

        return ColExpectation(col=self.col, operation=operation)

    def gt(self, other: int | float) -> Expectation:
        return OpComparisonExpectation(colname=self.col, value=other, operator=op.gt)

    def gte(self, other: int | float) -> Expectation:
        return OpComparisonExpectation(colname=self.col, value=other, operator=op.ge)

    def lt(self, other: int | float) -> Expectation:
        return OpComparisonExpectation(colname=self.col, value=other, operator=op.lt)

    def lte(self, other: int | float) -> Expectation:
        return OpComparisonExpectation(colname=self.col, value=other, operator=op.le)

    def equals(self, other: int | float) -> Expectation:
        return OpComparisonExpectation(colname=self.col, value=other, operator=op.eq)
    def non_null(self) -> Expectation:
        def operation(df: DataFrame) -> DataFrame:
            return df.withColumn("result", F.col(self.col).isNotNull())
        return ColExpectation(col=self.col, operation=operation)

class AllExpectationBuilder(Expectation):
    def __init__(self, *expectations: Expectation):
        self.expectations = expectations

    def run(self, dataframe_to_verify: "DataFrame") -> None:
        for expectation in self.expectations:
            expectation.run(dataframe_to_verify)


class AnyExpectationBuilder(Expectation):
    def __init__(self, *expectations: Expectation):
        self.expectations = expectations

    def run(self, dataframe_to_verify: "DataFrame") -> None:
        num_nonfailures = 0
        for expectation in self.expectations:
            try:
                expectation.run(dataframe_to_verify)
                num_nonfailures += 1
                return
            except AssertionError:
                pass
        if num_nonfailures == 0:
            raise AssertionError("No expectations passed")


any = AnyExpectationBuilder
all = AllExpectationBuilder
schema = SchemaBuilder
primary_key = PrimaryKeyExpectation
col = ColExpectationBuilder



from transforms.api.pipeline import Pipeline


def test_discovery():
    from . import example_module
    pipeline = Pipeline()
    pipeline.discover_transforms(example_module)
    
    assert len(pipeline._transforms) == 3


import re
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import TypedDict, TypeVar

import duckdb
from foundry_dev_tools import FoundryContext
from foundry_dev_tools.errors.dataset import BranchNotFoundError
from foundry_dev_tools.errors.meta import FoundryAPIError

from transforms.api import Transform

T = TypeVar("T")


@dataclass
class DatasetVersion:
    dataset_rid: str
    dataset_branch: str
    sanitized_rid: str
    sanitized_branch_name: str
    dataset_name:str
    dataset_identity: str
    last_update: datetime


class FoundryManager:
    def __init__(
        self,
        duckdb_conn: duckdb.DuckDBPyConnection,
        ctx: FoundryContext | None = None,
        branch_name: str | None = None,
        fallback_branches: list[str] | None = None,
    ):
        self.ctx = ctx or FoundryContext()
        self.duckdb_conn = duckdb_conn
        self.branch_name = branch_name or "master"
        if fallback_branches:
            self.fallback_branches = fallback_branches
        else:
            self.fallback_branches = [] if self.branch_name == "master" else ["master"]

        self.duckdb_conn.execute(
            f"CREATE SCHEMA IF NOT EXISTS fndry_{sanitize(self.branch_name)}"
        )
        self.duckdb_conn.execute("CREATE SCHEMA IF NOT EXISTS work")
        self.duckdb_conn.execute("CREATE SCHEMA IF NOT EXISTS meta")
        self.duckdb_conn.execute("""
            CREATE TABLE IF NOT EXISTS meta.datasets_versions (
                dataset_rid VARCHAR,
                dataset_branch VARCHAR,
                sanitized_rid VARCHAR,
                sanitized_branch_name VARCHAR,                 
                dataset_name VARCHAR,
                dataset_identity VARCHAR,
                last_update DATETIME
            ) """)

        for branch in self.fallback_branches:
            self.duckdb_conn.execute(
                f"CREATE SCHEMA IF NOT EXISTS fndry_{sanitize(branch)}"
            )
        self.duckdb_conn.commit()

    def collect_transform_inputs(self, transform: Transform) -> None:
        for input in transform.inputs.values():
            # Input has pinned branch - don't try to fallback
            if input.branch is not None:
                self.get_dataset_from_foundry_into_duckdb(
                    input.path_or_rid,
                    branch=input.branch,
                )
                
            else:
                try:
                    # Try main branch name
                    self.get_dataset_from_foundry_into_duckdb(
                        input.path_or_rid,
                        branch=self.branch_name,
                    )

                    
                except BranchNotFoundError as e:
                    for branch in self.fallback_branches:
                        # Try fallbacks and map back as view if found
                        created_dataset = self.get_dataset_from_foundry_into_duckdb(
                            input.path_or_rid,
                            branch=branch,
                        )
                        self.create_view(
                            src_schema=created_dataset["schema"],
                            src_table=created_dataset["tablename"],
                            target_schema=self.branch_name,
                            target_table=created_dataset["tablename"],
                        )
                        self.create_view(
                            src_schema=created_dataset["schema"],
                            src_table=created_dataset["tablename"],
                            target_schema=self.branch_name,
                            target_table=sanitize(input.path_or_rid),
                        )
                        break
                    else:
                        raise e

        return
    
    @contextmanager
    def download_file_to_temp_parquet(self,dataset_rid: str, branch: str ) :
        temp = tempfile.mkdtemp(suffix=f"foundry_dev_tools-{dataset_rid}")

        try: 
            self.ctx.cached_foundry_client.api.download_dataset_files(
                dataset_rid=dataset_rid,
                view=branch,
                output_directory=temp
            )
            yield temp
        except FoundryAPIError:
            yield "TODO"
            print("TODO: Download dataset through sql")
                
    def collect_transform_outputs(self, transform: Transform) -> None:
        return None

    def get_dataset_from_foundry_into_duckdb(
        self,
        dataset_rid: str,
        branch: str | None,
        update: bool = False,
    ) -> "DbDatasetInfo":
        branch_to_use = branch or self.branch_name
        identity = self.ctx.cached_foundry_client._get_dataset_identity(
            dataset_rid, branch=branch_to_use
        )
        meta = self.get_meta_for_dataset(dataset_rid, branch=branch_to_use)
        if meta and not update:
            return DbDatasetInfo(schema=sanitize(branch_to_use), tablename=dataset_rid)
        
        with self.download_file_to_temp_parquet(
            dataset_rid=dataset_rid,
            branch=branch_to_use,
        ) as temp_output:
            sanitized_rid = sanitize(dataset_rid)
            sanitized_dataset_name = sanitize(identity["dataset_path"].split("/")[-1])
            sanitized_branch_name = "fndry_" + sanitize(branch_to_use)
            self.duckdb_conn.execute(
                f"CREATE SCHEMA IF NOT EXISTS {sanitized_branch_name}"
            )
            temp_dataset_spark = Path(temp_output) / "spark/*"
            create_table_query = f"CREATE TABLE IF NOT EXISTS {sanitized_branch_name}.{sanitized_rid} AS SELECT * FROM read_parquet('{temp_dataset_spark}')"
            self.duckdb_conn.execute(create_table_query)

            self.duckdb_conn.execute(
                f"""INSERT INTO meta.datasets_versions by name (
                SELECT
                '{dataset_rid}' as dataset_rid,
                '{branch_to_use}' as dataset_branch,
                '{sanitized_rid}' as sanitized_rid,
                '{sanitized_branch_name}' as sanitized_branch_name,
                '{sanitized_rid}' as dataset_name,
                '{identity['last_transaction_rid']}' as dataset_identity,
                '{datetime.now()}' as last_update
            )
                """
            )
            self.create_view(
                src_schema=sanitized_branch_name,
                src_table=sanitized_rid,
                target_schema=sanitized_branch_name,
                target_table=sanitized_dataset_name,
            )
            
            return DbDatasetInfo(schema=sanitized_branch_name, tablename=dataset_rid)

    def create_view(
        self, src_schema: str, src_table: str, target_schema: str, target_table: str
    ):
        self.duckdb_conn.execute(f"""
            create or replace view {target_schema}.{target_table} as select * from {src_schema}.{src_table}
            """)

    def get_meta_for_dataset(self, dataset_rid: str, branch: str = "master"):
        res = self.duckdb_conn.query(
            f"SELECT * FROM meta.datasets_versions WHERE dataset_rid = '{dataset_rid}' AND dataset_branch = '{branch}'"
        )
        res1: (
            tuple[
                str,
                str,
                str,
                str,
                str,
                str,
                datetime,
            ]
            | None
        ) = res.fetchone()
        if res1:
            return DatasetVersion(*res1)


def sanitize(branch_name: str) -> str:
    return re.sub("[^a-zA-Z0-9_]", "_", branch_name).lower()


class DbDatasetInfo(TypedDict):
    schema: str
    tablename: str



import operator as op
from dataclasses import dataclass
from typing import Callable

from pyspark.sql import DataFrame

from .base import Expectation


@dataclass
class CountExpectation(Expectation):
    reference_count: int
    operator: Callable[[int, int], bool]

    def run(self, dataframe_to_verify: "DataFrame"):
        
        cnt = dataframe_to_verify.count()
        if cnt < self.reference_count: 
            raise AssertionError(f"Count is not {self.operator.__name__} than {self.reference_count}")

class CountExpectationBuilder:
    
    def gte(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.ge)
    def lte(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.le)
    def eq(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.eq)
    def equals(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.eq)
    def neq(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.ne)
    def gt(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.gt)
    def lt(self, value: int) -> CountExpectation:
        return CountExpectation(value, op.lt)

count = CountExpectationBuilder
    




import logging
from typing import Generator, TYPE_CHECKING

import pytest

from transforms.engine.duckdb import init_sess

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

@pytest.fixture(autouse=True)
def setup_logging():
    """Configure logging for tests"""
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

@pytest.fixture(scope="session")
def spark() -> "SparkSession":
    """Shared Spark-compatible session backed by DuckDB for all tests."""
    return init_sess()


@pytest.fixture(autouse=True)
def cleanup_spark(spark: "SparkSession") -> Generator[None, None, None]:
    """Cleanup Spark tables and views after each test."""
    yield
    for view in spark.catalog.listTables():
        spark.catalog.dropTempView(view.name)


from dataclasses import dataclass

from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from .base import Expectation


class GroupedExpectationBuilder:
    def __init__(self, *cols: str):
        self.cols = list(cols)

    def is_unique(self) -> Expectation:
        return GroupedExpectationIsUnique(self.cols)


@dataclass
class GroupedExpectationIsUnique(Expectation):
    cols: list[str]

    def run(self, dataframe_to_verify: "DataFrame"):
        res = (
            dataframe_to_verify.groupBy(*self.cols)
            .count()
            .filter(F.col("count") > 1)
            .collect()
        )
        if len(res) > 0:
            raise AssertionError(
                f"Grouped expectations are not unique, example issues: {res}"
            )


group_by = GroupedExpectationBuilder



def test_generate():
    from transforms.engine.duckdb import init_sess
    sess = init_sess()
    from transforms.generate_types import generate_from_spark
    parq = sess.read.parquet('transforms/tests/test_datasets/iris/spark/iris.parquet')
    generate_from_spark('iris',parq)
    pass



from transforms.engine.spark_sail import init_sess


def test_transform():
    sess = init_sess()
    parq = sess.read.parquet('transforms/tests/test_datasets/iris/spark/iris.parquet')
    count = parq.count()
    print(count)


def test_transform_big():
    sess = init_sess()
    def init_dataset():
        parq = sess.read.parquet('transforms/tests/test_datasets/iris/spark/iris.parquet')
        return parq
    df = init_dataset()
    for i in range(490):
        df = df.union(init_dataset())
    count = df.count()
    print(count)


from typing import TypeAlias, Literal, TYPE_CHECKING
if TYPE_CHECKING:
    from pyspark.sql import DataFrame




    ri_foundry_main_dataset_b0d1161d_5ae5_4e76_a6e2_9c7524eff80f:TypeAlias = 'DataFrame[Literal["id", "first_name", "last_name", "email", "gender", "ip_address", "comment", "image", "dt"]]'





    test2_95:TypeAlias = 'DataFrame[Literal["id", "first_name", "last_name", "email", "gender", "ip_address", "comment", "image", "dt"]]'



from typing import TYPE_CHECKING, cast, Any

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


def _patch_sqlframe_writer_save() -> None:
    """
    Runtime patch for SQLFrame DuckDB writer to support df.write.format(...).save(...)
    and df.write.save(...).
    """
    try:
        from sqlframe.duckdb.readwriter import DuckDBDataFrameWriter  # type: ignore[import]
        from sqlframe.base.readerwriter import _infer_format  # type: ignore[import]
    except Exception:
        # If sqlframe is not available, or the import shape changes, fail gracefully.
        return

    if hasattr(DuckDBDataFrameWriter, "save"):
        # Newer sqlframe versions might already support this API; nothing to do.
        return

    def save(
        self: Any,
        path: str,
        format: str | None = None,
        mode: str | None = None,
        **options: Any,
    ) -> None:
        # Prefer explicit format, then writer.state_format_to_write, then infer from path.
        fmt = format or getattr(self, "_state_format_to_write", None)
        if fmt is None:
            fmt = _infer_format(path)

        effective_mode = mode or getattr(self, "_mode", None)

        # Delegate to the engine-specific _write implementation
        self._write(path=path, mode=effective_mode, format=fmt, **options)  # type: ignore[call-arg]

    setattr(DuckDBDataFrameWriter, "save", save)


def init_sess() -> "SparkSession":
    """Initialize a DuckDB session that mimics the SparkSession interface."""
    from sqlframe import activate

    activate(engine="duckdb")
    _patch_sqlframe_writer_save()

    from pyspark.sql import SparkSession

    sess = cast("SparkSession", SparkSession.builder.appName("test").getOrCreate())
    return sess


from typing import Iterable


def test_patch_rdd():
    from sqlframe import activate
    activate(engine="duckdb")
    from pyspark.sql import Row, SparkSession

    from transforms.runner.patches.rdd import patch_dataframe_rdd

    sess = SparkSession.builder.appName("test").getOrCreate()  # type: ignore
    sess: SparkSession
    df = sess.createDataFrame([(1, 2), (3, 4)], "a int, b int")
    df = patch_dataframe_rdd(df)
    def some_func(rows:Iterable[Row]):
        for row in rows:
            new_data = row.asDict()
            new_data['a'] = new_data['a'] + 1
            
            yield Row(**new_data)
    df2 = df.rdd.mapPartitions(some_func)
    assert df2 != df
    df2.toDF("someschema")


import tempfile
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import alembic.config
import duckdb
from foundry_dev_tools import FoundryContext
from foundry_dev_tools.errors.dataset import BranchNotFoundError
from foundry_dev_tools.errors.meta import FoundryAPIError
from sqlmodel import Session, create_engine, select

from .models import DatasetIdentifier


@dataclass
class DatasetVersion:
    dataset_rid: str
    dataset_branch: str
    sanitized_rid: str
    sanitized_branch_name: str
    dataset_name: str
    dataset_identity: str
    last_update: datetime


class DataManager:
    def __init__(
        self,
        metadata_path: Path = Path.home() / ".fndry_duck" / "meta.db",
        storage_dir: Path = Path.home() / ".fndry_duck" / "store",
        config_dir: Path = Path.home() / ".fndry_duck",
        ctx: FoundryContext | None = None,
    ):
        self._engine = create_engine(f"sqlite:///{str(metadata_path)}")
        self.session = Session(self._engine)
        self.metadata_path = metadata_path
        self.storage_dir = storage_dir
        self.config_dir = config_dir
        self.ctx = ctx or FoundryContext()
        # cli.run(
        #     ["migrate", "deploy", f'--schema={Path(__file__).parent/'schema.prisma'}'],
        #     env={"DATABASE_URL": self.metadata_path},
        # )

        alembicArgs = [
            '--raiseerr',
            'upgrade', 'head',
        ]
        alembic.config.main(argv=alembicArgs)

    def reset_meta(self):
        self.metadata_path.unlink(missing_ok=True)

        alembicArgs = [
            '--raiseerr',
            'upgrade', 'head',
        ]
        alembic.config.main(argv=alembicArgs)
        
        
        


        

    async def load_latest_parquet_into_database(
        self,
        dataset_rid_or_path: str,
        branch: str,
    ):
        self.session.exec(select(DatasetIdentifier).where(DatasetIdentifier.rid_or_path == dataset_rid_or_path))
        dataset = await self.prisma.dataset_identifier.find_first(
            where={"rid_or_path": dataset_rid_or_path}, include={"dataset": True}
        )
        if dataset:
            version = await self.prisma.dataset_version.find_first(
                where={
                    "branch": {"is": {"full_branch_name": "branch"}},
                    "datasetId": dataset.id,
                },
                order={"data_identity_date": "desc"},
            )

        identity = self.ctx.cached_foundry_client._get_dataset_identity(
            dataset_path_or_rid=dataset_rid_or_path, branch=branch
        )
        identity_id = identity["last_transaction_rid"]
        # No version found or is outdated and there is some version available
        if identity_id is None:
            return "No data"
        if not version or (identity_id != version.data_identity_id):
            temp = tempfile.mkdtemp(
                suffix=f"foundry_dev_tools-{identity['last_transaction_rid']}"
            )

            try:
                self.ctx.cached_foundry_client.api.download_dataset_files(
                    dataset_rid=identity_id,
                    view=branch,
                    output_directory=temp,
                )
                return temp
            except FoundryAPIError:
                print("TODO: Download dataset through sql")
                return "TODO"

        else:
            return identity_id



from dataclasses import dataclass

from transforms.runner.data_source.base import BranchNotFoundError, DataSource
import logging
logger = logging.getLogger(__name__)

@dataclass
class MixedDataSource(DataSource):
    sources: dict[str, DataSource] 
    fallback_source: DataSource | None
    
    async def download_dataset(self, dataset_path_or_rid: str, branch: str):
        source = self.sources.get(branch)
        if source is None:
            if self.fallback_source is None:
                raise BranchNotFoundError('SOURCE')
            source = self.fallback_source
        return await source.download_dataset(dataset_path_or_rid, branch=branch)

        

    async def download_for_branches(self, dataset_path_or_rid: str, branches: list[str]):
        for branch in branches:
            try:
                return await self.download_dataset(dataset_path_or_rid, branch=branch)

            except BranchNotFoundError as e:
                logger.info(f"[{e.source}] Branch [{branch}] not found for dataset [{dataset_path_or_rid}]")
        raise BranchNotFoundError(source="MIXED")




from abc import ABC, abstractmethod

from pyspark.sql import DataFrame


class Expectation(ABC):
    @abstractmethod
    def run(self, dataframe_to_verify: "DataFrame") -> None:
        raise NotImplementedError("Expectation has to be implemented")


from .check import Check
from .configure import configure
from .incremental_transform import incremental
from .pipeline import Pipeline
from .transform_df import (
    Input,
    Output,
    Transform,
    TransformInput,
    TransformOutput,
    transform,
    transform_df,
)

__all__ = [
    "Input",
    "Output",
    "Transform",
    "transform_df",
    "Check",
    "configure",
    "incremental",
    "transform",
    "TransformInput",
    "TransformOutput",
    "Pipeline",
]



from dataclasses import dataclass
from pathlib import Path

from pyspark.sql import DataFrame

from transforms.runner.data_sink.base import DataSink


@dataclass
class LocalFileSink(DataSink):
    branch: str
    output_dir: str = str((Path.home() / ".fndry_duck" / "local_output"))

    def save_transaction(
        self,
        df: DataFrame,
        dataset_path_or_rid: str,
    ) -> None:
        result_path = Path(self.output_dir) / self.branch / dataset_path_or_rid
        
        df.write.parquet(
            str(result_path),
            mode="overwrite",
            compression="zstd",
        )



import pytest
from pathlib import Path
from typing import Dict, List, Any, cast
from pyspark.sql import DataFrame, SparkSession, Row
from transforms.api.transform_df import Transform, TransformOutput, Input, Output
from transforms.runner.data_sink.base import DataSink
from transforms.runner.data_source.base import DataSource
from transforms.runner.exec_transform import TransformRunner
from transforms.engine.duckdb import init_sess

class MockDataSource(DataSource):
    def __init__(self, data_dict: Dict[str, DataFrame]):
        self.data = data_dict
    
    async def download_for_branches(self, dataset_path_or_rid: str, branches: List[str]) -> DataFrame:
        df = self.data.get(dataset_path_or_rid)
        if df is None:
            raise ValueError(f"Dataset {dataset_path_or_rid} not found")
        return df
    
    async def download_latest_incremental_transaction(self, dataset_path_or_rid: str, branches: List[str], semantic_version: int) -> DataFrame:
        df = self.data.get(f"{dataset_path_or_rid}_incremental")
        if df is None:
            raise ValueError(f"Incremental dataset {dataset_path_or_rid} not found")
        return df

class MockDataSink(DataSink):
    def __init__(self):
        self.saved_data: Dict[str, DataFrame] = {}
        self.saved_incremental: Dict[str, DataFrame] = {}
    
    def save_transaction(self, df: DataFrame, dataset_path_or_rid: str) -> None:
        self.saved_data[dataset_path_or_rid] = df
    
    def save_incremental_transaction(self, df: DataFrame, dataset_path_or_rid: str, semantic_version: int) -> None:
        self.saved_incremental[f"{dataset_path_or_rid}_{semantic_version}"] = df

def simple_transform(input1: DataFrame, input2: DataFrame) -> DataFrame:
    return input1.union(input2)

def multi_transform(input1: DataFrame, output1: TransformOutput, output2: TransformOutput) -> None:
    df1 = input1.filter("value % 2 = 0")
    df2 = input1.filter("value % 2 = 1")
    output1.on_dataframe_write(df1, "replace")
    output2.on_dataframe_write(df2, "replace")
    return None

class SimpleTransform(Transform):
    def __init__(self):
        inputs = {
            "input1": Input(path_or_rid="test_input1"),
            "input2": Input(path_or_rid="test_input2")
        }
        outputs = {
            "output": Output(path_or_rid="test_output", checks=[])
        }
        super().__init__(inputs=inputs, outputs=outputs, transform=simple_transform)

class MultiOutputTransform(Transform):
    def __init__(self):
        inputs = {
            "input1": Input(path_or_rid="test_input1")
        }
        outputs = {
            "output1": Output(path_or_rid="test_output1", checks=[]),
            "output2": Output(path_or_rid="test_output2", checks=[])
        }
        super().__init__(inputs=inputs, outputs=outputs, transform=multi_transform)
        self.multi_outputs = {}

@pytest.fixture
def sample_data(spark: SparkSession) -> Dict[str, DataFrame]:
    data1 = [(i,) for i in range(5)]
    data2 = [(i,) for i in range(5, 10)]
    df1 = spark.createDataFrame(data1, ["value"])
    df2 = spark.createDataFrame(data2, ["value"])
    return {
        "test_input1": df1,
        "test_input2": df2,
        "test_input1_incremental": df1.limit(2)
    }

@pytest.fixture
def transform_runner(sample_data: Dict[str, DataFrame]) -> TransformRunner:
    source = MockDataSource(sample_data)
    sink = MockDataSink()
    return TransformRunner(
        sourcer=source,
        sink=sink,
        fallback_branches=["main"],
        output_dir=Path("/tmp/test_output"),
        secrets_config_location=Path("/tmp/test_secrets")
    )

@pytest.mark.asyncio
async def test_simple_transform(transform_runner: TransformRunner, sample_data: Dict[str, DataFrame]) -> None:
    transform = SimpleTransform()
    sources = await transform_runner.download_datasets(transform, omit_checks=True, dry_run=False)
    transform_runner.exec_transform(transform, omit_checks=True, dry_run=False, sources=sources)
    
    # Verify the output
    output_df = transform_runner.sink.saved_data["test_output"]
    expected_count = (
        sample_data["test_input1"].count() + 
        sample_data["test_input2"].count()
    )
    assert output_df.count() == expected_count

@pytest.mark.asyncio
async def test_multi_output_transform(transform_runner: TransformRunner, sample_data: Dict[str, DataFrame]) -> None:
    transform = MultiOutputTransform()
    sources = await transform_runner.download_datasets(transform, omit_checks=True, dry_run=False)
    
    # Debug: Print sources before execution
    print(f"Sources before execution: {sources.keys()}")
    print(f"Transform outputs: {transform.outputs.keys()}")
    print(f"Transform multi_outputs: {transform.multi_outputs}")
    
    transform_runner.exec_transform(transform, omit_checks=True, dry_run=False, sources=sources)
    
    # Debug: Print saved data after execution
    print(f"Saved data keys: {transform_runner.sink.saved_data.keys()}")
    
    # Verify the outputs
    output1_df = transform_runner.sink.saved_data["test_output1"]
    output2_df = transform_runner.sink.saved_data["test_output2"]
    
    assert output1_df.count() == 3  # Even numbers
    assert output2_df.count() == 2  # Odd numbers
    
    even_values = [row["value"] for row in output1_df.collect()]
    odd_values = [row["value"] for row in output2_df.collect()]
    
    assert all(v % 2 == 0 for v in even_values)
    assert all(v % 2 == 1 for v in odd_values)

@pytest.mark.asyncio
async def test_dry_run(transform_runner: TransformRunner) -> None:
    transform = SimpleTransform()
    sources = await transform_runner.download_datasets(transform, omit_checks=True, dry_run=True)
    transform_runner.exec_transform(transform, omit_checks=True, dry_run=True, sources=sources)
    
    # Verify no data was saved
    assert len(transform_runner.sink.saved_data) == 0
    assert len(transform_runner.sink.saved_incremental) == 0 



from pathlib import Path
from pyspark.sql import DataFrame
def generate_from_spark(dataset_name: str,spark_df: DataFrame):
    types_path = Path(__file__).parent / "types.py"
    
    with open(types_path, "r") as f:
        dataset = f.readlines()
        
    with open(types_path, "w") as f:
        replaced_declaration = False
        names =', '.join([f'"{name}"' for name in spark_df.columns])
        declaration = f"\n    {dataset_name}:TypeAlias = 'DataFrame[Literal[{names}]]'\n"
        for line in dataset:
            if line.startswith(f'    {dataset_name}'):
                replaced_declaration = True
                f.write(declaration)
            else:
                
                f.write(line)
        if not replaced_declaration:
            f.write(declaration)
        
            
    return




from typing import Optional
from pyspark.sql import SparkSession


def init_sess(sail_server_url: Optional[str] = None) -> SparkSession:
    
    default_sail_server_url = "localhost:50051"
    
    spark = SparkSession.builder.remote(
        f"sc://{sail_server_url or default_sail_server_url}"
    ).getOrCreate()

    return spark



from dataclasses import dataclass
from typing import Literal

from ..expectations.colexpects import Expectation


@dataclass
class Check:
    expectation: Expectation
    description: str
    on_error: Literal['FAIL', "WARN"] = "WARN"


from dataclasses import dataclass
from pathlib import Path

from pyspark.sql import SparkSession

from transforms.runner.data_source.base import BranchNotFoundError, DataSource
import logging
logger = logging.getLogger(__name__)


@dataclass
class LocalDataSource(DataSource):
    session: SparkSession
    output_dir: str = str((Path.home() / ".fndry_duck" / "local_output"))

    async def download_dataset(self, dataset_path_or_rid: str, branch: str):
        if not (Path(self.output_dir) / branch / dataset_path_or_rid).exists():
            raise BranchNotFoundError('LOCAL')

        return self.session.read.parquet(
            f"{self.output_dir}/{branch}/{dataset_path_or_rid}/*.parquet"
        )

    async def download_for_branches(self, dataset_path_or_rid: str, branches: list[str]):
        for branch in branches:
            try:
                return await self.download_dataset(dataset_path_or_rid, branch=branch)

            except BranchNotFoundError:
                logger.info(f"[LOCAL]Branch [{branch}] not found for dataset [{dataset_path_or_rid}]")
        raise BranchNotFoundError(source='LOCAL')



from pyspark.sql import DataFrame

from ..api.check import Check


def execute_check(df: DataFrame, check: Check ):
    try:
        check.expectation.run(df)
        print(f"Check {check.description} passed successfully")
    except Exception as e:
        print(f"{check.on_error}: Check {check.description} failed with error {e}")
        if check.on_error=='FAIL':
            raise e

    



from typing_extensions import Self
from pyspark.sql import DataFrame
from .base import Expectation



class ConditionalExpectationBuilder(Expectation):
    def __init__(self, when_pairs: list[tuple[Expectation, Expectation]]):
        self.when_pairs = when_pairs
        self.otherwise_expr: Expectation | None = None
        
    def when(self, when_expr: Expectation, then_expr: Expectation) -> Self:
        self.when_pairs.append((when_expr, then_expr))

        return self
    def otherwise(self, otherwise_expr: Expectation) -> Self:
        self.otherwise_expr = otherwise_expr
        return self
    def run(self, dataframe_to_verify: "DataFrame") -> None:
        # TODO: implement
        pass

        

class ConditionalExpectationStart:
    def __init__(self, when_expr: Expectation,
    then_expr: Expectation):
        ConditionalExpectationStart(when_expr, then_expr)

    

when = ConditionalExpectationStart



from typing import Literal


def edit_annotations(transform_path: str, to_annotate: dict[str, str], mode: Literal['add', 'remove']) -> None:
    # TODO: Implement it
    # to_annotate = mapping[rid, type_name in transforms.types]
    # 1. Read the file
    # 2. Find line starting transform_df decorator
    # 2a. go through arguments and extract rids of inputs
    # 3. Find line starting decorated function
    # 3a. add/remove types according to mode
    # from typing import TYPE_CHECKING
    # if TYPE_CHECKING:
    #     from transforms.types import 
    # def compute(df1: "dt.sometype"):
    # ...
    raise NotImplementedError()




from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, Union, Dict, Optional, Any
from asyncio import gather
from asyncer import syncify
from pyspark.sql import DataFrame

from transforms.api.transform_df import Transform, TransformOutput
from transforms.runner.data_sink.base import DataSink
from transforms.runner.data_source.base import DataSource

from .exec_check import execute_check
import logging
logger = logging.getLogger(__name__)

@dataclass
class TransformRunner:
    sourcer: DataSource
    sink: DataSink
    fallback_branches: list[str] = field(default_factory=list)
    output_dir: Path = Path.home() / ".fndry_duck" / "output"
    secrets_config_location: Path = Path.home() / ".fndry_duck" / "secrets"
    
    def __post_init__(self):
        self.output_dir.mkdir(parents=True, exist_ok=True)

    async def download_datasets(self, transform: Transform, omit_checks: bool, dry_run: bool) -> Dict[str, Union[DataFrame, TransformOutput]]:
        sources: Dict[str, Union[DataFrame, TransformOutput]] = {}
        futures_list = [(argname, self.sourcer.download_for_branches(input.path_or_rid, branches=[
                b for b in ([input.branch] + self.fallback_branches) if b is not None
            ])) for argname, input in transform.inputs.items()]
        futures = [fut for (_, fut) in futures_list]
        names = [name for (name, _) in futures_list]
        dfs: list[DataFrame] = await gather(*futures)
        for argname, res_df in zip(names, dfs):
            sources[argname] = res_df
        return sources
            
    def exec_transform(self, transform: Transform, omit_checks: bool, dry_run: bool, sources: Optional[Dict[str, Union[DataFrame, TransformOutput]]] = None) -> None:
        if sources is None:
            sources = syncify(self.download_datasets, raise_sync_error=False)(transform, omit_checks, dry_run)
            
        if transform.multi_outputs is not None:
            impl_multi_outputs = {}
            for argname, output in transform.outputs.items():
                def on_dataframe_req(mode: Literal["current", "previous"]) -> DataFrame:
                    if mode == "current":
                        return syncify(self.sourcer.download_for_branches)(
                            output.path_or_rid, branches=self.fallback_branches
                        )
                    elif (transform.incremental_opts is not None) and (not dry_run):
                        return syncify(self.sourcer.download_latest_incremental_transaction)(
                            dataset_path_or_rid=output.path_or_rid,
                            branches=self.fallback_branches,
                            semantic_version=transform.incremental_opts.semantic_version
                        )
                    raise ValueError(f"Invalid mode {mode} or missing incremental options")

                # Create a closure to capture the current output variable
                def create_write_fn(output_path: str):
                    def on_dataframe_write(df: DataFrame, mode: Literal["append", "replace"]) -> None:
                        if mode == "append":
                            raise NotImplementedError()
                        else:
                            if not omit_checks:
                                for check in output.checks: 
                                    execute_check(df, check)
                            if (transform.incremental_opts is not None) and (not dry_run):
                                self.sink.save_incremental_transaction(
                                    df,
                                    output_path,
                                    transform.incremental_opts.semantic_version
                                )
                            else:
                                logger.info(f"Saving output to {output_path}")
                                self.sink.save_transaction(
                                    df=df,
                                    dataset_path_or_rid=output_path
                                )
                    return on_dataframe_write

                output_df_impl = TransformOutput(
                    on_dataframe_req=on_dataframe_req,
                    on_dataframe_write=create_write_fn(output.path_or_rid)
                )
                impl_multi_outputs[argname] = output_df_impl
                sources[argname] = output_df_impl
            
            # Set the multi_outputs dictionary with our implementations
            transform.multi_outputs = impl_multi_outputs
        
        logger.info("Starting transform")
        res = transform.transform(**sources)
        logger.info("Finished transform")
        
        if transform.multi_outputs is None:
            # Accept any DataFrame-like object (including engine-specific SQLFrame DataFrames)
            if not hasattr(res, "limit") or not hasattr(res, "collect"):
                raise ValueError("Transform without multi_outputs must return a DataFrame-like object")
            
            if not omit_checks:
                for check in transform.outputs["output"].checks:
                    logger.info(f"Running check {check.description}")
                    execute_check(res if not dry_run else res.limit(1), check)
                    logger.info(f"Check {check.description} finished")
            if not dry_run:
                logger.info(f"Saving transaction to {transform.outputs['output'].path_or_rid}")
                self.sink.save_transaction(df=res, dataset_path_or_rid=transform.outputs['output'].path_or_rid)
                logger.info("Transaction saved successfully")
            else:
                res.limit(1).collect()
        else:
            if transform.incremental_opts is not None:
                logger.info("Finished transform successfully")
                raise NotImplementedError("Not yet implemented saving for incremental dataset")
            else:
                logger.info("Finished transform successfully")
                



import sys
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
)
logger = logging.getLogger(__name__)

class Engine(str, Enum):
    spark = "spark"
    duckdb = "duckdb"
    sparksail = "spark-sail"


def find_path_where_there_is_setup(module_name: str) -> str:
    parent = Path(module_name).parent
    files = parent.glob("setup.py")
    if str(parent) == module_name:
        raise Exception("Root reached, no setup.py found")
    if len(list(files)) == 0:
        return find_path_where_there_is_setup(str(parent))
    else:
        return str(parent)


@contextmanager
def traverse_to_setup_and_add_to_path(module_name: str):
    path = find_path_where_there_is_setup(module_name=module_name)
    sys.path.insert(0, str(path))
    try:
        sys.path.insert(0, str(path))
        yield
    finally:
        sys.path.remove(str(path))


if __name__ == "__main__":

    def main(
        transform_to_run: str,
        fallback_branches: str,
        omit_checks: Annotated[
            bool, typer.Option(help="Disables checks running")
        ] = False,
        transform_name: Annotated[
            Optional[str],
            typer.Option(
                "--transform-name",
                help="Name of the transform function in the module to execute "
                     "(required if the module defines multiple transforms)",
            ),
        ] = None,
        engine: Annotated[
            Engine,
            typer.Option(help="Engine to use for the transformation"),
        ] = Engine.spark,
        sail_server_url: Annotated[
            Optional[str], typer.Option(help="Sail server url")
        ] = None,
        dry_run: Annotated[
            bool, typer.Option(help="Dry run the transformation")
        ] = False,
        local_dev_branch_name: Annotated[
            str, typer.Option(help="Branch name for local development")
        ] = "duck-fndry-dev",
    ):
        with traverse_to_setup_and_add_to_path(transform_to_run):
            logger.info(f"Starting engine {engine}")
            if engine == engine.duckdb:
                from transforms.engine.duckdb import init_sess

                session = init_sess()
            if engine == Engine.sparksail:
                from transforms.engine.spark_sail import init_sess

                session = init_sess(sail_server_url)
            else:
                from transforms.engine.spark import init_sess

                session = init_sess()
            logger.info(f"Started engine {engine}")
            from .runner.default_executor import execute_with_default_foundry

            execute_with_default_foundry(
                dry_run=dry_run,
                fallback_branches=fallback_branches,
                transform_name=transform_name,
                session=session,
                local_dev_branch_name=local_dev_branch_name,
                omit_checks=omit_checks,
                transform_to_run=transform_to_run,
            )

    typer.run(main)



from dataclasses import dataclass
from pathlib import Path
from typing import Callable

import duckdb
from pyspark.sql import DataFrame

from ...generate_types import generate_from_spark

from .local_file_sink import LocalFileSink


@dataclass
class LocalFileSinkWithDuck(LocalFileSink):
    duckdb_path: str = str((Path.home() / ".fndry_duck" / "analytical_db.db"))
    get_dataset_dataset_name: Callable[[str], str] = lambda x: x

    def __post_init__(self):
        Path(self.duckdb_path).parent.mkdir(parents=True, exist_ok=True)

    def save_transaction(
        self,
        df: DataFrame,
        dataset_path_or_rid: str,
    ) -> None:
        result_path = Path(self.output_dir) / self.branch / dataset_path_or_rid
        result_path.mkdir(parents=True, exist_ok=True)
        super().save_transaction(
            df=df,
            dataset_path_or_rid=str(Path(dataset_path_or_rid) / "data.parquet"),
        )
        self.conn = duckdb.connect(self.duckdb_path, config={})
        dataset_name = self.get_dataset_dataset_name(dataset_path_or_rid)
        generate_from_spark(dataset_name, df)
        self.conn.execute(
            f"CREATE OR REPLACE VIEW {dataset_name} as select * from read_parquet('{self.output_dir}/{self.branch}/{dataset_path_or_rid}/*.parquet')"
        )

    def save_incremental_transaction(
        self,
        df: DataFrame,
        dataset_path_or_rid: str,
        semantic_version: int,
    ):
        # TODO: Implement
        pass



from typing import Any

from transforms.engine.duckdb import init_sess


def test_duckdb_sqlframe_supports_writer_format_save() -> None:
    """
    Ensure that our runtime patch enables df.write.format(...).save(...)
    and df.write.save(...) (with format inference).

    This exercises the patched API lightly to guard against regressions
    in SQLFrame/DuckDB integration.
    """
    sess = init_sess()

    df = sess.range(0, 3).toDF("id")  # type: ignore[attr-defined]

    # Use a temporary path in the current working directory.
    base_path = "tmp_duckdb_format_save_test"

    # This should not raise for explicit format
    df.write.format("parquet").save(f"{base_path}.parquet")  # type: ignore[call-arg]

    # And this should not raise with format inferred from the path
    df.write.save(f"{base_path}.csv")  # type: ignore[call-arg]




from abc import ABC
from dataclasses import dataclass

from pyspark.sql import DataFrame


@dataclass
class DataSource(ABC):
    async def download_dataset(
        self,
        dataset_path_or_rid: str,
        branch: str,
    ) -> DataFrame:
        raise NotImplementedError()

    async def download_for_branches(
        self, dataset_path_or_rid: str, branches: list[str]
    ) -> DataFrame:
        raise NotImplementedError()

    def get_last_transaction(
        self, dataset_path_or_rid: str, branches: list[str]
    ) -> DataFrame:
        raise NotImplementedError()

    async def download_latest_incremental_transaction(
        self, dataset_path_or_rid: str, branches: list[str], semantic_version: int
    ) -> "DataFrame":
        # TODO: Implement it
        raise NotImplementedError()


@dataclass
class BranchNotFoundError(Exception):
    source: str



import importlib.util
import sys

from typing import Any

from foundry_dev_tools import FoundryContext
from pyspark.sql import SparkSession

from transforms.api.transform_df import Transform
from transforms.runner.exec_transform import TransformRunner

from .data_sink.local_file_sink_with_duck import LocalFileSinkWithDuck
from .data_source.base import DataSource
from .data_source.foundry_source_with_duck import FoundrySourceWithDuck
from .data_source.local_file_source import LocalDataSource
from .data_source.mixed_source import MixedDataSource


def execute_with_default_foundry(
    transform_to_run: str,
    fallback_branches: str,
    omit_checks: bool,
    session: SparkSession,
    dry_run: bool,
    local_dev_branch_name: str,
    transform_name: str | None = None,
):
    mod = import_from_path("transform", transform_to_run)
    transforms: dict[str, Transform | Any] = {}
    for name, item in mod.__dict__.items():
        if isinstance(item, Transform):
            transforms[name] = item

    if not transforms:
        print("file has no transforms")
        return

    if transform_name is not None:
        if transform_name not in transforms:
            print(
                f"Transform '{transform_name}' not found in module. "
                f"Available transforms: {list(transforms.keys())}"
            )
            return
        selected_transform: Transform = transforms[transform_name]  # type: ignore[assignment]
    else:
        if len(transforms) > 1:
            print("There is more than one transform specified. Please specify its name.")
            print("names are", list(transforms.keys()))
            return
        selected_transform = list(transforms.values())[0]  # type: ignore[assignment]

    branches = fallback_branches.split(",")
    all_branches = [local_dev_branch_name] + branches


    foundry_source = FoundrySourceWithDuck(
        ctx=FoundryContext(), session=session, 
    )
    local_source = LocalDataSource(session=session)
    sources_mapping: dict[str, DataSource] = {b: foundry_source for b in branches}
    sources_mapping[local_dev_branch_name] = local_source

    TransformRunner(
        sink=LocalFileSinkWithDuck(
            branch=local_dev_branch_name, 
        ),
        sourcer=MixedDataSource(
            sources=sources_mapping,
            fallback_source=foundry_source,
        ),
        fallback_branches=all_branches,
    ).exec_transform(
        selected_transform,
        omit_checks=omit_checks,
        dry_run=dry_run,
    )


def import_from_path(module_name: str, file_path: str):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    sys.modules[module_name] = module  # Register the module in sys.modules
    spec.loader.exec_module(  # type:ignore
        module
    )  # Execute the module in its own namespace
    return module



import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from transforms.api.transform_df import Transform


@dataclass
class Source:
    external_system_rid: str
    secrets_config_location: str | None = None

    def get_secret(self, secret_name: str)->str:
        if self.secrets_config_location is None:
            raise ValueError("No secrets config location specified")
        with open(self.secrets_config_location, 'r') as f:
            secrets = json.load(f)
            system= secrets.get(self.external_system_rid)
            if not system:
                raise ValueError(f"No secrets for external system {self.external_system_rid}")
            key = system.get(secret_name)
            if not key:
                raise ValueError(f"No secret for key {key} in external system {self.external_system_rid}")
            return key
        raise NotImplementedError()

def external_systems(**kwargs: Source):
    def _external_systems(transform: "Transform"):
        transform.external_systems = kwargs
        return transform
    return _external_systems


from hashlib import sha256
import re
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import duckdb

from ...generate_types import generate_from_spark

from .foundry_source import FoundrySource

if TYPE_CHECKING:
    from pyspark.sql import DataFrame


@dataclass
class FoundrySourceWithDuck(FoundrySource):
    duckdb_path: str = str((Path.home() / ".fndry_duck" / "analytical_db.db"))
    duckdb_path_sql: str = str((Path.home() / ".fndry_duck" / "analytical_db.sql"))
    # get_dataset_dataset_name: Callable[[str], str] = lambda x: x

    
    def get_dataset_name(self,dataset_path_or_rid: str) -> str:
        dataset_path = str(self.ctx.get_dataset(dataset_path_or_rid).path)
        sha_addon = sha256(dataset_path.encode()).hexdigest()[:2]
        dataset_name = dataset_path.split("/")[-1]
        return f"{dataset_name}_{sha_addon}"

    async def download_dataset(
        self, dataset_path_or_rid: str, branch: str
    ) -> "DataFrame":
        df = await super().download_dataset(dataset_path_or_rid, branch)
        self.conn = duckdb.connect(self.duckdb_path)
        sanitized_branch = re.sub("[^0-9a-zA-Z]+", "_", branch)

        self.conn.execute(f"create schema if not exists {sanitized_branch}")
        dataset_name = self.get_dataset_name(dataset_path_or_rid)
        generate_from_spark(dataset_name, df)

        if self.last_path.endswith(".parquet"):
            self.conn.execute(
                f"CREATE OR REPLACE VIEW {sanitized_branch}.{dataset_name} as select * from read_parquet('{self.last_path}/**/*.parquet')"
            )
        elif self.last_path.endswith(".csv"):
            self.conn.execute(
                f"CREATE OR REPLACE VIEW {sanitized_branch}.{dataset_name} as select * from read_csv('{self.last_path}/*.csv')"
            )
        else:
            raise NotImplementedError(
                f"Format {self.last_path.split('.')[-1]} is not supported"
            )
        with open(self.duckdb_path_sql, "w") as f:
            for row in self.conn.query(
                f"select schema_name from information_schema.schemata where catalog_name = '{Path(self.duckdb_path).stem}'"
            ).fetchall():
                f.write(f"create schema if not exists {row[0]};\n")
            for row in self.conn.query(
                "select sql from duckdb_views() where not internal"
            ).fetchall():
                qry: str = row[0]
                view_replacement = qry.replace(
                    "CREATE VIEW ", "CREATE OR REPLACE VIEW "
                )
                f.write(f"{view_replacement};\n")
        self.conn.close()
        return df

    def download_latest_incremental_transaction(
        self, dataset_path_or_rid: str, branches: list[str], semantic_version: int
    ) -> "DataFrame":
        # TODO: Implement it
        raise NotImplementedError()



import os
from pathlib import Path

import pytest
from dotenv import load_dotenv
from foundry_dev_tools import FoundryContext, JWTTokenProvider
from foundry_dev_tools.errors.config import MissingCredentialsConfigError

from transforms.engine.duckdb import init_sess
from transforms.runner.data_source.foundry_source_with_duck import FoundrySourceWithDuck


# Load environment variables from a .env file (if present)
# This allows configuring e2e test parameters via a local .env in the repo root.
load_dotenv()

FOUNDRY_E2E_ENABLED = os.getenv("FOUNDRY_DUCK_E2E_ENABLED") == "1"
FOUNDRY_E2E_DATASET = os.getenv("FOUNDRY_DUCK_E2E_DATASET_PATH", "somerid")
FOUNDRY_E2E_BRANCH = os.getenv("FOUNDRY_DUCK_E2E_BRANCH", "master")
FOUNDRY_E2E_HOST = os.getenv("FOUNDRY_HOST", "somehost.com")
FOUNDRY_E2E_TOKEN = os.getenv("FOUNDRY_TOKEN", "somejwt.com")


pytestmark = pytest.mark.e2e



@pytest.mark.asyncio
@pytest.mark.skipif(
    not FOUNDRY_E2E_ENABLED or not FOUNDRY_E2E_DATASET or not FOUNDRY_E2E_HOST or not FOUNDRY_E2E_TOKEN,
    reason="Foundry e2e tests not configured (set FOUNDRY_DUCK_E2E_ENABLED=1 and FOUNDRY_DUCK_E2E_DATASET_PATH)",
)
async def test_foundry_source_with_duck_downloads_real_dataset(tmp_path: Path) -> None:
    """
    Basic end-to-end test that:
    - connects to a real Foundry environment via foundry-dev-tools
    - downloads a dataset into a DuckDB-backed SQLFrame session
    - verifies that the resulting DataFrame is usable
    - verifies that DuckDB view definitions are materialized

    This test requires:
    - valid foundry-dev-tools configuration in the environment
    - FOUNDRY_DUCK_E2E_ENABLED=1
    - FOUNDRY_DUCK_E2E_DATASET_PATH set to a readable dataset path or RID
    - optionally FOUNDRY_DUCK_E2E_BRANCH (defaults to 'master')
    """

    try:
        ctx = FoundryContext(token_provider=JWTTokenProvider(jwt=FOUNDRY_E2E_TOKEN, host=FOUNDRY_E2E_HOST))
    except MissingCredentialsConfigError as exc:
        pytest.skip(
            f"Foundry credentials not configured for e2e tests: {exc}. "
            "See https://emdgroup.github.io/foundry-dev-tools/getting_started/installation.html"
        )
    session = init_sess()

    duckdb_db_path = tmp_path / "analytical_db.db"
    duckdb_sql_path = tmp_path / "analytical_db.sql"

    source = FoundrySourceWithDuck(
        ctx=ctx,
        session=session,
        duckdb_path=str(duckdb_db_path),
        duckdb_path_sql=str(duckdb_sql_path),
        
    )

    df = await source.download_dataset(FOUNDRY_E2E_DATASET, branch=FOUNDRY_E2E_BRANCH)
    

    # The returned object should be a SQLFrame DataFrame (Spark-like API)
    assert hasattr(df, "limit")
    assert hasattr(df, "collect")

    # Ensure we can touch the data without errors
    df.limit(1).collect()
    assert df.count() > 5

    # Ensure DuckDB artifacts were written alongside the download
    assert duckdb_db_path.exists()
    assert duckdb_sql_path.exists()
    assert duckdb_sql_path.read_text().strip() != ""

